#include <UrsSTM32F411-fast-ADC.h>

#if __has_include( "Config.h" ) // Der STM-Compliler untersttzt C++14
#include "Config.h"
#endif

static ADC_TypeDef& adcDevice = *ADC1;
static ADC_Common_TypeDef& adcCommon = *ADC1_COMMON;

// ADC1 kann ber DMA2, Stream0/Channel0 oder Stream4/Channel0 ausgelesen werden.
static DMA_TypeDef& dmaController = *DMA2;
static DMA_Stream_TypeDef& dmaStream = *DMA2_Stream0;

static UrsDmaISR pTransferComplete = 0;
static UrsDmaISR pHalfTransferComplete = 0;
static UrsAdcISR pAdcIrq = 0;

static volatile int conversionCount = 0; // Anzahl durchgefhrter Konvertierungen

UrsAdcResolution UrsAdcClass::currentResolution = UrsAdcResolution::Bit12;

void UrsAdcClass::enable(bool waitForStabilisation) {
   volatile uint32_t counter = 0U;

   RCC->APB2ENR |= RCC_APB2ENR_ADC1EN; // Enable ADC Clock

   // Set to reset state
   adcCommon.CCR = 0;
   adcDevice.SR = 0;
   adcDevice.CR1 = 0;
   adcDevice.CR2 = 0;
   adcDevice.SMPR1 = 0;
   adcDevice.SMPR2 = 0;
   adcDevice.JOFR1 = 0;
   adcDevice.JOFR2 = 0;
   adcDevice.JOFR3 = 0;
   adcDevice.JOFR4 = 0;
   adcDevice.HTR = 0x00000FFF;
   adcDevice.LTR = 0;
   adcDevice.SQR1 = 0;
   adcDevice.SQR2 = 0;
   adcDevice.SQR3 = 0;
   adcDevice.JSQR = 0;

   adcDevice.CR2 = ADC_CR2_ADON;  // Enable ADC

   currentResolution = UrsAdcResolution::Bit12;
   HAL_NVIC_DisableIRQ(ADC_IRQn);

   if (waitForStabilisation) {
      // Delay for ADC stabilization time
      // Compute number of CPU cycles to wait for */
      counter = (ADC_STAB_DELAY_US * (SystemCoreClock / 1'000'000U));
      while (counter != 0U) {
         counter--;
      }
   }
}

void UrsAdcClass::disable() {
   HAL_NVIC_DisableIRQ(ADC_IRQn);
   adcDevice.CR2 &= ~ADC_CR2_ADON; // disable ADC
   RCC->APB2ENR &= ~RCC_APB2ENR_ADC1EN; // disable ADC Clock
}

void UrsAdcClass::setClockDivider(UrsAdcClockDivider cd) {
   adcCommon.CCR = (adcCommon.CCR & ~ADC_CCR_ADCPRE_Msk) | (uint32_t)cd;
}

void UrsAdcClass::setResolution(UrsAdcResolution r) {
   currentResolution = r;
   adcDevice.CR1 = (adcDevice.CR1 & ~ADC_CR1_RES_Msk) | (uint32_t)r;
}

void UrsAdcClass::setEocSelection(UrsAdcEocSelection es) {
   adcDevice.CR2 = (adcDevice.CR2 & ~ADC_CR2_EOCS_Msk) | (uint32_t)es;
}

/*
void UrsAdcClass::setDataAlign(UrsAdcDataAlign da) {
   adcDevice.CR2 = (adcDevice.CR2 & ~ADC_CR2_ALIGN_Msk) | (uint32_t)da;
}
*/
/*
void UrsAdcClass::setTrigger(UrsAdcTrigger t) {
   adcDevice.CR2 = (adcDevice.CR2 & ~ADC_CR2_EXTSEL_Msk) | (uint32_t)t;
}
*/

void UrsAdcClass::setChannelSampleTime(UrsAdcChannel channel, UrsAdcSampleTime sampleTime) {
   // if ADC_Channel_10 ... ADC_Channel_18 is selected
   if ((uint32_t)channel > ADC_CHANNEL_9) {
      // Clear the old sample time
      adcDevice.SMPR1 &= ~ADC_SMPR1(ADC_SMPR1_SMP10, (uint32_t)channel);

      // Set the new sample time
      adcDevice.SMPR1 |= ADC_SMPR1((int)sampleTime, (uint32_t)channel);
   }
   else { // ADC_Channel include in ADC_Channel_[0..9]
      // Clear the old sample time
      adcDevice.SMPR2 &= ~ADC_SMPR2(ADC_SMPR2_SMP0, (uint32_t)channel);

      // Set the new sample time
      adcDevice.SMPR2 |= ADC_SMPR2((int)sampleTime, (uint32_t)channel);
   }
}

void UrsAdcClass::setSequence(std::initializer_list<UrsAdcChannel> channelList) {
   int count = 0;

   int size = channelList.size();
   if (size > 16)
      size = 16;

   // Set ADC number of conversion to size
   adcDevice.SQR1 = (adcDevice.SQR1 & ~ADC_SQR1_L_Msk) | ADC_SQR1(size);

   for (UrsAdcChannel channel : channelList) {
      addChannelToSequence(++count, channel);
      if (count > size)
         break;
   }
}

void UrsAdcClass::addChannelToSequence(uint32_t rank, UrsAdcChannel channel) {
   // For Rank 1 to 6
   if (rank < 7U) {
      // Clear the old SQx bits for the selected rank
      adcDevice.SQR3 &= ~ADC_SQR3_RK(ADC_SQR3_SQ1, rank);

      // Set the SQx bits for the selected rank
      adcDevice.SQR3 |= ADC_SQR3_RK(channel, rank);
   }
   // For Rank 7 to 12
   else if (rank < 13U) {
      // Clear the old SQx bits for the selected rank
      adcDevice.SQR2 &= ~ADC_SQR2_RK(ADC_SQR2_SQ7, rank);

      // Set the SQx bits for the selected rank
      adcDevice.SQR2 |= ADC_SQR2_RK(channel, rank);
   }
   // For Rank 13 to 16
   else {
      // Clear the old SQx bits for the selected rank
      adcDevice.SQR1 &= ~ADC_SQR1_RK(ADC_SQR1_SQ13, rank);

      // Set the SQx bits for the selected rank
      adcDevice.SQR1 |= ADC_SQR1_RK(channel, rank);
   }

   // if adcDevive Channel_18 is selected for VBAT Channel ennable VBATE
   if (channel != UrsAdcChannel::Vbat) {
      setGpioToAdcFunction((PinName)channel, PinMap_ADC);
   }
   else {
      // Disable the TEMPSENSOR channel in case of using board with multiplixed ADC_CHANNEL_VBAT & ADC_CHANNEL_TEMPSENSOR
      if ((uint16_t)ADC_CHANNEL_TEMPSENSOR == (uint16_t)ADC_CHANNEL_VBAT) {
         adcCommon.CCR &= ~ADC_CCR_TSVREFE;
      }
      // Enable the VBAT channel
      adcCommon.CCR |= ADC_CCR_VBATE;
   }
}

void UrsAdcClass::setGpioToAdcFunction(PinName pin, const PinMap* map) {
   while (map->pin != NC) {
      if (map->pin == pin) {
         pin_function(pin, map->function);
         return;
      }
      map++;
   }
}

uint16_t UrsAdcClass::analogRead(UrsAdcChannel channel, UrsAdcSampleTime sampleTime) {
   setSequence({ channel });

   setChannelSampleTime(channel, sampleTime);

   adcDevice.SR = 0;

   adcDevice.CR2 |= (uint32_t)ADC_CR2_SWSTART;
   while (!(adcDevice.SR & ADC_FLAG_EOC)) // Warten auf Ende der Konvertierung
      ;

   return adcDevice.DR & 0xFFFF;
}

void UrsAdcClass::setReadChannel(UrsAdcChannel channel, UrsAdcSampleTime sampleTime) {
   setChannelSampleTime(channel, sampleTime);
   setSequence({ channel });
}

uint16_t UrsAdcClass::repeatRead() {
   adcDevice.SR = 0;

   adcDevice.CR2 |= (uint32_t)ADC_CR2_SWSTART;
   while (!(adcDevice.SR & ADC_FLAG_EOC)) // Warten auf Ende der Konvertierung
      ;

   return adcDevice.DR & 0xFFFF;
}

// ==============================================
// Read sequnces with DMA
// ==============================================

#ifdef INSTALL_DMA_ISR
extern "C" {
   // DMA ISR
   void DMA2_Stream0_IRQHandler() {
      if (dmaController.LISR & DMA_LISR_TCIF0) { // Transfer complete
         if (pAdcIrq)
            HAL_NVIC_DisableIRQ(ADC_IRQn);

         adcDevice.CR1 &= ~ADC_CR1_SCAN; // Scan-Mode abstellen.
         adcDevice.CR2 &= ~(ADC_CR2_DMA | ADC_CR2_CONT); // DMA und kontinuierlichen Modus abstellen.

         dmaStream.CR &= ~DMA_SxCR_EN; // DMA stop
         // Der HTC wird bei abgeschaltetem HTIE zusammen mit dem TC ausgelst
         dmaController.LIFCR |= DMA_LIFCR_CTCIF0 | DMA_LIFCR_CHTIF0; // Flag lschen
         if (pTransferComplete)
            pTransferComplete();
      }

      if (dmaController.LISR & DMA_LISR_HTIF0) { // Half transfer complete
         dmaController.LIFCR |= DMA_LIFCR_CHTIF0; // Flag lschen
         pHalfTransferComplete();
      }
   }
}
#endif // NO_DMA_ISR


#ifdef INSTALL_ADC_ISR
extern "C" {
   void  ADC_IRQHandler() {
      if (pAdcIrq)
         pAdcIrq(conversionCount - dmaStream.NDTR);
   }
}
#endif // !NO_ADC_ISR

void UrsAdcClass::initDma(void* dest, uint16_t count, UrsDmaISR transferComplete, UrsDmaISR halfTransferComplete) {
   __HAL_RCC_DMA2_CLK_ENABLE(); // DMA controller clock enable

   // Check ADC DMA Mode
   // - disable the DMA Mode if it is already enabled
   if ((adcDevice.CR2 & ADC_CR2_DMA) == ADC_CR2_DMA) {
      adcDevice.CR2 &= ~ADC_CR2_DMA;
   }

   dmaStream.CR &= ~DMA_SxCR_EN; // DMA stop 
   while (dmaStream.CR & DMA_SxCR_EN) // Wait until stopped
      ;

   dmaStream.CR = 0 // Set configuration register
      | (0 << DMA_SxCR_CHSEL_Pos)  // Channel selection: channel 0
      | (0 << DMA_SxCR_MBURST_Pos) // Memory burst transfer configuration: single transfer
      | (0 << DMA_SxCR_PBURST_Pos) // Peripheral burst transfer configuration: single transfer
      | (0 << DMA_SxCR_CT_Pos)     // Current target: not used
      | (0 << DMA_SxCR_DBM_Pos)    // Double buffer mode: No buffer switching
      | (0 << DMA_SxCR_PL_Pos)     // Priority level: low
      | (0 << DMA_SxCR_PINCOS_Pos) // Peripheral increment offset size: 0
      | (0 << DMA_SxCR_MSIZE_Pos)  // Memory data size: byte (8-bit)
      | (0 << DMA_SxCR_PSIZE_Pos)  // Peripheral data size: byte (8-bit)
      | (1 << DMA_SxCR_MINC_Pos)   // Memory increment mode: Memory address pointer is incremented
      | (0 << DMA_SxCR_PINC_Pos)   // Peripheral increment mode: fixed
      | (0 << DMA_SxCR_CIRC_Pos)   // Circular mode: disabled
      | (0 << DMA_SxCR_DIR_Pos)    // Data transfer direction: Peripheral-to-memory
      | (0 << DMA_SxCR_PFCTRL_Pos) // Peripheral flow controller: The DMA is the flow controller
      | (0 << DMA_SxCR_TCIE_Pos)   // Transfer complete interrupt: enable
      | (0 << DMA_SxCR_HTIE_Pos)   // Half transfer complete interrupt: disable
      | (0 << DMA_SxCR_TEIE_Pos)   // Transfer error interrupt: disable
      | (0 << DMA_SxCR_DMEIE_Pos)  // Direct mode error interrupt: disable
      ;

   if ((currentResolution == UrsAdcResolution::Bit10) || (currentResolution == UrsAdcResolution::Bit12)) {
      dmaStream.CR |=
         (1 << DMA_SxCR_MSIZE_Pos)  // Memory data size: half-word (16-bit)
         | (1 << DMA_SxCR_PSIZE_Pos);  // Peripheral data size: half-word (16-bit)
   }

#ifdef INSTALL_DMA_ISR
   pTransferComplete = transferComplete;
   pHalfTransferComplete = halfTransferComplete;
   dmaStream.CR |= DMA_SxCR_TCIE; // Transfer complete interrupt: enable

   if (halfTransferComplete) {
      dmaStream.CR |= DMA_SxCR_HTIE; // Half transfer complete interrupt: enable
   }
#endif // NO_DMA_ISR



   dmaStream.PAR = (uint32_t)&adcDevice.DR; // Data source: adcDevive data register
   dmaStream.M0AR = (uint32_t)dest;         // Destination address
   dmaStream.FCR = 0;                       // FIFO not used
   dmaStream.NDTR = count;
   conversionCount = count;

   // Clear interrupt flags
   dmaController.LIFCR |= DMA_LIFCR_CTCIF0 | DMA_LIFCR_CHTIF0 | DMA_LIFCR_CTEIF0 | DMA_LIFCR_CDMEIF0 | DMA_LIFCR_CFEIF0;

   // DMA interrupt init
   HAL_NVIC_SetPriority(DMA2_Stream0_IRQn, 0, 0); // DMA2_Stream0_IRQn interrupt configuration
   HAL_NVIC_EnableIRQ(DMA2_Stream0_IRQn);

   dmaStream.CR |= DMA_SxCR_EN;     // DMA enable
}

void UrsAdcClass::startSequence(std::initializer_list<UrsAdcChannel> channelList, uint16_t* dest,
   uint16_t count
#ifdef INSTALL_DMA_ISR
   , UrsDmaISR transferComplete, UrsDmaISR halfTransferComplete
#endif // NO_DMA_ISR
#ifdef INSTALL_ADC_ISR
   , UrsAdcISR adcConversionDone
#endif // NO_ADC_ISR
) {

#ifdef INSTALL_DMA_ISR
   initDma(dest, channelList.size() * count, transferComplete, halfTransferComplete);
#else
   initDma(dest, channelList.size() * count, nullptr, nullptr);
#endif // NO_DMA_ISR

   setSequence(channelList);

   // ADCStarten
   adcDevice.SR = 0;

   adcDevice.CR1 |= ADC_CR1_SCAN; // Scan-Mode einstellen.

#ifdef INSTALL_ADC_ISR
   pAdcIrq = adcConversionDone;

   

   if (adcConversionDone) {
      adcDevice.CR1 |= ADC_CR1_EOCIE; // EOC-Interrupt enable.

      // ADC interrupt init
      HAL_NVIC_SetPriority(ADC_IRQn, 0, 0); // ADC_IRQn interrupt configuration
      HAL_NVIC_EnableIRQ(ADC_IRQn);
   }
#endif // NO_ADC_ISR

   adcDevice.CR2 |= ADC_CR2_DMA | ADC_CR2_CONT; // DMA und kontinuierlichen Modus einstellen.

   adcDevice.CR2 |= (uint32_t)ADC_CR2_SWSTART; // ADC starten
}

bool UrsAdcClass::isConversionComplete() {
   return dmaStream.NDTR == 0;
}

int UrsAdcClass::getConversionsDone() {
   return conversionCount - dmaStream.NDTR;
}

UrsAdcClass UrsADC;