#pragma once

#include <Arduino.h>
#include <UrsSTM32F411-fast-ADC-Definitions.h>
#include <initializer_list>

#if __has_include( "UrsADC-Config.h" ) // Der STM-Compliler untersttzt C++14
#include "UrsADC-Config.h"
#endif

// Prototypdefinition fr die ISR (Transfer Complete und Half Transfer Complete).
typedef void (*UrsDmaISR)();
typedef void (*UrsAdcISR)(int count); // Liefert die Anzahl der bereits abgetasteten Sequenzen.

extern "C" void DMA2_Stream0_IRQHandler(void); // DMA ISR 

class UrsAdcClass {
   friend void DMA2_Stream0_IRQHandler(void);
public:
   static void enable(bool waitForStabilisation = true);  // Initialisiert den ADC
   static void disable(); // Stoppt den ADC
   static void setClockDivider(UrsAdcClockDivider); // Legt den Frequenzteiler fest
   static void setResolution(UrsAdcResolution);     // Legt die Auflsung fest 6..12 Bit
   static void setEocSelection(UrsAdcEocSelection);

   // static void setDataAlign(UrsAdcDataAlign);    // Legt die Datenausrichtung fest
   // static void setTrigger(UrsAdcTrigger);

   // Fhrt eine einzelne Konvertierung durch.
   // Wie: Arduino uint32_t analogRead(uint32_t ulPin).
   static uint16_t analogRead(UrsAdcChannel channel, UrsAdcSampleTime sampleTime = UrsAdcSampleTime::Cycles_3);

   // Stellt den zu lesenden Kanal fr repeatRead ein.
   static void setReadChannel(UrsAdcChannel channel, UrsAdcSampleTime sampleTime = UrsAdcSampleTime::Cycles_3);
   // Den von setReadChannel eingestellten Kanal erneut auslesen.
   static uint16_t repeatRead();

   // Legt die kanalspezifische Abtastzeit fest.
   static void setChannelSampleTime(UrsAdcChannel channel, UrsAdcSampleTime sampleTime);

   // Startet die mehrfache Konvertierung einer Sequenz.
   // channelList:      Liste der Kanle einer einzelnen Sequenz [1..16].
   // dest:             Zieladresse, die ermittelten Daten werden dorthin kopiert.
   //                   ! Es werden conversionList.size * count * 1 bzw. 2 Bytes kopiert.
   // count:            Anzahl Wiederholungen der Sequenz.
   //                   ! conversionList.size * count darf 65.535 nicht berschreiten.
   // transferComplete: ISR fr den Transfer-Complete-Interrupt.
   // halfTransferComplete: ISR fr den Half-Transfer-Complete-Interrupt.
   // adcConversionDone: ISR fr ADC Interrupt
   static void startSequence(std::initializer_list<UrsAdcChannel> channelList,
      uint16_t* dest, uint16_t count = 1
#ifdef INSTALL_DMA_ISR
      , UrsDmaISR transferComplete = nullptr, UrsDmaISR halfTransferComplete = nullptr
#endif // !NO_DMA_ISR
#ifdef INSTALL_ADC_ISR
      , UrsAdcISR adcConversionDone = nullptr
#endif // !NO_ADC_ISR

   );

   // Die bei startSequence eingestellte Anzahl Konvertierungen wurde ausgefhrt.
   static bool isConversionComplete();

   // Liefert die Anzahl der nach Ausfhrung von startSequence bereits durchgefhrten Konvertierungen.
   static int getConversionsDone();

protected:
   static UrsAdcResolution currentResolution;
   // Legt die Auslese-Sequenz fest. Max 16 Kanle sind mglich.
   static void setSequence(std::initializer_list<UrsAdcChannel> channel);

   // Fgt der Sequenz einen Kanal hinzu.
   // rank: Postition in der Sequenz [1..16]
   static void addChannelToSequence(uint32_t rank, UrsAdcChannel channel);

   // Der angegebene Pin wird auf Funktion ADC-Input eingestellt.
   static void setGpioToAdcFunction(PinName pin, const PinMap* map);

   // Initialisiert die DMA
   // dest: Zieladresse
   // count: Anzahl zu bertragender Daten
   // transferComplete: ISR fr Transfer Complete Interrupt. nullptr, wenn nicht genutzt.
   // halfTransferComplete: ISR fr Half Transfer Complete Interrupt. nullptr, wenn nicht genutzt.
   static void initDma(void* dest, uint16_t count, UrsDmaISR transferComplete, UrsDmaISR halfTransferComplete);
};

extern UrsAdcClass UrsADC;